import argparse
import atexit
import errno
import logging
import os

import uvicorn

import phoenix.config as config
from phoenix.server.app import create_app

logger = logging.getLogger(__name__)


def _write_pid_file() -> None:
    with open(_get_pid_file(), "w"):
        pass


def _remove_pid_file() -> None:
    try:
        os.unlink(_get_pid_file())
    except OSError as e:
        if e.errno == errno.ENOENT:
            # If the pid file doesn't exist, ignore and continue on since
            # we are already in the desired end state; This should not happen
            pass
        else:
            raise


def _get_pid_file() -> str:
    return os.path.join(config.get_pids_path(), "%d" % os.getpid())


if __name__ == "__main__":
    # automatically remove the pid file when the process is being gracefully terminated
    atexit.register(_remove_pid_file)
    _write_pid_file()

    parser = argparse.ArgumentParser()
    parser.add_argument("--primary", type=str)
    parser.add_argument("--reference", type=str)
    parser.add_argument("--port", type=int, default=config.port)
    parser.add_argument("--graphiql", action="store_true")
    parser.add_argument("--debug", action="store_true")
    args = parser.parse_args()

    # Validate the required args
    if args.primary is None:
        raise ValueError("Primary dataset is required via the --primary flag")
    if args.reference is None:
        raise ValueError("Reference dataset is required via the --reference flag")
    print(
        f"""Starting Phoenix App
            primary dataset: {args.primary}
            reference dataset: {args.reference}"""
    )

    app = create_app(
        primary_dataset_name=args.primary,
        reference_dataset_name=args.reference,
        debug=args.debug,
        graphiql=args.graphiql,
    )

    uvicorn.run(app, port=args.port)



from typing import Optional

import strawberry
from strawberry.types import Info

from phoenix.server.api.context import Context


@strawberry.type
class DimensionDataQuality:
    dimension_name: strawberry.Private[str]

    @strawberry.field
    async def cardinality(self, info: Info[Context, None]) -> Optional[int]:
        return await info.context.loaders.cardinality.load(self.dimension_name)



from itertools import chain
from typing import List

from pandas import DataFrame

from . import errors as err
from .schema import Schema


def validate_dataset_inputs(dataframe: DataFrame, schema: Schema) -> List[err.ValidationError]:
    general_checks = chain(check_missing_columns(dataframe, schema))
    return list(general_checks)


def check_missing_columns(dataframe: DataFrame, schema: Schema) -> List[err.MissingColumns]:
    # converting to a set first makes the checks run a lot faster
    existing_columns = set(dataframe.columns)
    missing_columns = []

    for field in schema.__dict__:
        if field.endswith("column_name"):
            col = getattr(schema, field)
            if col is not None and col not in existing_columns:
                missing_columns.append(col)

    if schema.feature_column_names is not None:
        for col in schema.feature_column_names:
            if col not in existing_columns:
                missing_columns.append(col)

    if schema.embedding_feature_column_names is not None:
        for emb_col_names in schema.embedding_feature_column_names.values():
            if emb_col_names.vector_column_name not in existing_columns:
                missing_columns.append(emb_col_names.vector_column_name)
            if (
                emb_col_names.raw_data_column_name is not None
                and emb_col_names.raw_data_column_name not in existing_columns
            ):
                missing_columns.append(emb_col_names.raw_data_column_name)
            if (
                emb_col_names.link_to_data_column_name is not None
                and emb_col_names.link_to_data_column_name not in existing_columns
            ):
                missing_columns.append(emb_col_names.link_to_data_column_name)

    if missing_columns:
        return [err.MissingColumns(missing_columns)]
    return []



from typing import Dict, List

from pandas.api.types import is_numeric_dtype, is_object_dtype

from phoenix.datasets import Dataset

from .dimension import Dimension
from .dimension_data_type import DimensionDataType
from .dimension_type import DimensionType


class Model:
    def __init__(self, primary_dataset_name: str, reference_dataset_name: str):
        # TODO Fail if you can't find the datasets on disc
        self.__primary_dataset = Dataset.from_name(primary_dataset_name)
        self.__reference_dataset = Dataset.from_name(reference_dataset_name)
        self.__dimensions = self._get_dimensions(self.primary_dataset, self.reference_dataset)
        # TODO construct embedding dimensions from the dataset schemas
        self.__embedding_dimensions: List[Dimension] = []

    @property
    def primary_dataset(self) -> Dataset:
        return self.__primary_dataset

    @property
    def reference_dataset(self) -> Dataset:
        return self.__reference_dataset

    @property
    def dimensions(self) -> List[Dimension]:
        return self.__dimensions

    @property
    def embedding_dimensions(self) -> List[Dimension]:
        return self.__embedding_dimensions

    def _get_dimensions(
        self, primary_dataset: Dataset, reference_dataset: Dataset
    ) -> List[Dimension]:
        # TODO: include reference dataset dimensions
        dimensions: List[Dimension] = []
        schema = primary_dataset.schema

        actual_column_names = [
            name
            for name in [
                schema.actual_label_column_name,
                schema.actual_score_column_name,
            ]
            if name is not None
        ]
        prediction_column_names = [
            name
            for name in [
                schema.prediction_label_column_name,
                schema.prediction_score_column_name,
            ]
            if name is not None
        ]
        feature_column_names = (
            schema.feature_column_names if schema.feature_column_names is not None else []
        )
        tag_column_names = schema.tag_column_names if schema.tag_column_names is not None else []
        dimension_type_to_column_names: Dict[DimensionType, List[str]] = {
            DimensionType.ACTUAL: actual_column_names,
            DimensionType.PREDICTION: prediction_column_names,
            DimensionType.FEATURE: feature_column_names,
            DimensionType.TAG: tag_column_names,
        }

        for dimension_type, column_names in dimension_type_to_column_names.items():
            for name in column_names:
                dimensions.append(
                    Dimension(
                        name=name,
                        data_type=self._infer_dimension_data_type(name),
                        type=dimension_type,
                    )
                )

        return dimensions

    @staticmethod
    def _get_embedding_dimensions(
        primary_dataset: Dataset, reference_dataset: Dataset
    ) -> List[Dimension]:
        raise NotImplementedError()

    def _infer_dimension_data_type(self, dimension_name: str) -> DimensionDataType:
        # TODO: verify corresponding dimension of reference dataset has same type
        dimension_pandas_dtype = self.primary_dataset.dataframe[dimension_name].dtype
        if is_numeric_dtype(dimension_pandas_dtype):  # type: ignore
            return DimensionDataType.NUMERIC
        elif is_object_dtype(dimension_pandas_dtype):  # type: ignore
            return DimensionDataType.CATEGORICAL
        raise ValueError("Unrecognized dimension type")



from dataclasses import dataclass
from typing import Optional, Union

from starlette.requests import Request
from starlette.responses import Response
from starlette.websockets import WebSocket

from phoenix.core.model import Model

from .loaders import Loaders


@dataclass
class Context:
    request: Union[Request, WebSocket]
    response: Optional[Response]
    model: Model
    loaders: Loaders



import strawberry

from phoenix.core import Dimension as CoreDimension

from .DimensionDataQuality import DimensionDataQuality
from .DimensionDataType import DimensionDataType
from .node import Node


@strawberry.type
class Dimension(Node):
    name: str
    dataType: DimensionDataType

    @strawberry.field
    def dataQuality(self) -> DimensionDataQuality:
        return DimensionDataQuality(dimension_name=self.name)


def to_gql_dimension(id_attr: int, dimension: CoreDimension) -> Dimension:
    """
    Converts a phoenix.core.Dimension to a phoenix.server.api.types.Dimension
    """
    return Dimension(
        id_attr=id_attr,
        name=dimension.name,
        dataType=DimensionDataType[dimension.data_type.value],
    )



import os
from typing import Optional, Union

from starlette.applications import Starlette
from starlette.requests import Request
from starlette.responses import Response
from starlette.routing import Mount, Route, WebSocketRoute
from starlette.staticfiles import StaticFiles
from starlette.types import Scope
from starlette.websockets import WebSocket
from strawberry.asgi import GraphQL
from strawberry.schema import BaseSchema

from phoenix.core.model import Model

from .api.context import Context
from .api.loaders import Loaders, create_loaders
from .api.schema import schema


class Static(StaticFiles):
    "Static file serving with a fallback to index.html"

    async def get_response(self, path: str, scope: Scope) -> Response:

        response = await super().get_response(path, scope)
        print("code", response.status_code)
        if response.status_code == 404:
            full_path, stat_result = self.lookup_path("index.html")
            if stat_result is None:
                raise RuntimeError("Filed to find index.html")
            return self.file_response(full_path, stat_result, scope)

        return response


class GraphQLWithContext(GraphQL):
    def __init__(
        self, schema: BaseSchema, model: Model, loader: Loaders, graphiql: bool = False
    ) -> None:
        self.model = model
        self.loader = loader
        super().__init__(schema, graphiql=graphiql)

    async def get_context(
        self,
        request: Union[Request, WebSocket],
        response: Optional[Response] = None,
    ) -> Context:

        return Context(request=request, response=response, model=self.model, loaders=self.loader)


def create_app(
    primary_dataset_name: str,
    reference_dataset_name: str,
    debug: bool = False,
    graphiql: bool = False,
) -> Starlette:
    model = Model(
        primary_dataset_name=primary_dataset_name, reference_dataset_name=reference_dataset_name
    )
    graphql = GraphQLWithContext(
        schema=schema, model=model, loader=create_loaders(model), graphiql=graphiql
    )
    return Starlette(
        debug=debug,
        routes=[
            Route(
                "/graphql",
                graphql,
            ),
            WebSocketRoute("/graphql", graphql),
            Mount(
                "/",
                app=Static(
                    directory=os.path.join(
                        os.path.dirname(__file__),
                        "static",
                    ),
                    html=True,
                ),
                name="static",
            ),
        ],
    )



import errno
import os
import tempfile


def normalize_path(path: str) -> str:
    """Normalizes the given path by converting it to an absolute path and
    expanding the user directory, if necessary.
    Args:
        path: a path
    Returns:
        the normalized path
    """
    return os.path.expanduser(path)


def _get_temp_path() -> str:
    """Get path to  directory in which to store temp phoenix server files."""
    return os.path.join(tempfile.gettempdir(), ".arize-phoenix")


def get_pids_path() -> str:
    """Get path to directory in which to store temp phoenix instance pid files.
    This directory is used to track any currently running instances of Arize Phoenix
    on the host machine. The directory will be created if it does not exist.
    """
    path = os.path.join(_get_temp_path(), "pids")
    try:
        os.makedirs(path)
    except OSError as e:
        if e.errno == errno.EEXIST:
            pass
        else:
            raise
    else:
        os.chmod(path, 0o777)
    return path


PHOENIX_DIR = os.path.dirname(os.path.abspath(__file__))
ROOT_DIR = os.path.join("~", ".phoenix")
dataset_dir = normalize_path(os.path.join(ROOT_DIR, "datasets"))

# Server config
server_dir = os.path.join(PHOENIX_DIR, "server")
# The port the server will run on after launch_app is called
port = 6060



import logging
from typing import Optional

import phoenix.config as config
from phoenix.datasets import Dataset
from phoenix.services import AppService

try:
    from IPython.display import IFrame  # type: ignore
except:  # noqa
    pass

logger = logging.getLogger(__name__)


_session: Optional["Session"] = None


class Session:
    "Session that maintains a 1-1 shared state with the Phoenix App."

    def __init__(self, primary: Dataset, reference: Dataset, port: int):
        self.primary = primary
        self.reference = reference
        self.port = port
        # Initialize an app service that keeps the server running
        self._app_service = AppService(port, primary.name, reference.name)
        self._is_colab = _is_colab()

    def view(self) -> "IFrame":
        # Display the app in an iframe
        return IFrame(src=_get_url(self.port, self._is_colab), width="100%", height=1000)

    def end(self) -> None:
        "Ends the session and closes the app service"
        self._app_service.stop()


# TODO(mikeldking): validate that we really want to require a reference dataset. Leaving it optional
# Provides a level of flexibility
def launch_app(primary: Dataset, reference: Dataset) -> "Session":
    "Launches the phoenix application"
    logger.info("Launching Phoenix App")
    global _session

    _session = Session(primary, reference, port=config.port)

    return _session


def close_app() -> None:
    "Closes the phoenix application"
    global _session
    if _session is None:
        print("No active session to close")
        return
    _session.end()
    _session = None
    logger.info("Session closed")


def _get_url(port: int, is_colab: bool) -> str:
    """Determines the iframe url based on whether this is in a Colab"""
    if is_colab:
        from google.colab.output import eval_js  # type: ignore

        return str(eval_js(f"google.colab.kernel.proxyPort({port}, {{'cache': true}})"))

    return f"http://localhost:{port}/"


def _is_colab() -> bool:
    """Determines whether this is in a Colab"""
    try:
        import google.colab  # type: ignore # noqa: F401
        import IPython  # type: ignore
    except ImportError:
        return False

    return IPython.get_ipython() is not None



import logging
import os
import signal
import subprocess
import sys
from typing import List

import psutil

import phoenix.config as config

logger = logging.getLogger(__name__)


class Service:
    """Interface for phoenix services.
    All services must define a ``command`` property.
    """

    working_dir = "."

    def __init__(self) -> None:
        self.child = self.start()

    @property
    def command(self) -> List[str]:
        raise NotImplementedError(f"{type(self)} must define `command`")

    def start(self) -> psutil.Popen:
        """Starts the service."""

        if len(os.listdir(config.get_pids_path())) > 0:
            # Currently, only one instance of Phoenix can be running at any given time.
            # Support for multiple concurrently running instances may be supported in the future.
            logger.warning(
                "Existing running Phoenix instance detected! Shutting "
                "it down and starting a new instance..."
            )
            Service.stop_any()

        return psutil.Popen(
            self.command,
            cwd=self.working_dir,
            stdin=subprocess.PIPE,
            env={**os.environ},
        )

    def stop(self) -> None:
        """Gracefully stops the service."""
        self.child.stdin.close()
        try:
            self.child.wait(timeout=5)
        except psutil.TimeoutExpired:
            self.child.terminate()

    @staticmethod
    def stop_any() -> None:
        """Stops any running instance of the service, whether the instance is being run
        within the current session or if it is being run in a separate process on the
        same host machine. In either case, the instance will be forcibly stopped.
        """
        pids_path = config.get_pids_path()
        for filename in os.listdir(pids_path):
            os.kill(int(filename), signal.SIGKILL)
            filename_path = os.path.join(pids_path, filename)
            os.unlink(filename_path)


class AppService(Service):
    """Service that controls the phoenix application."""

    working_dir = config.server_dir

    def __init__(self, port: int, primary_dataset_name: str, reference_dataset_name: str):
        self.port = port
        self.__primary_dataset_name = primary_dataset_name
        self.__reference_dataset_name = reference_dataset_name
        super().__init__()

    @property
    def command(self) -> List[str]:

        command = [
            sys.executable,
            "main.py",
            "--port",
            str(self.port),
            "--primary",
            str(self.__primary_dataset_name),
            "--reference",
            str(self.__reference_dataset_name),
        ]
        logger.info(f"command: {command}")
        return command



from typing import Optional

import strawberry
from strawberry.arguments import UNSET
from strawberry.types import Info

from phoenix.server.api.context import Context

from .Dimension import Dimension, to_gql_dimension
from .pagination import Connection, ConnectionArgs, Cursor, connection_from_list


@strawberry.type
class Model:
    @strawberry.field
    def dimensions(
        self,
        info: Info[Context, None],
        first: Optional[int] = 50,
        last: Optional[int] = None,
        after: Optional[Cursor] = UNSET,
        before: Optional[Cursor] = UNSET,
    ) -> Connection[Dimension]:
        """
        A non-trivial implementation should efficiently fetch only
        the necessary books after the offset.
        For simplicity, here we build the list and then slice it accordingly
        """
        return connection_from_list(
            [
                to_gql_dimension(index, dimension)
                for index, dimension in enumerate(info.context.model.dimensions)
            ],
            args=ConnectionArgs(
                first=first,
                after=after if isinstance(after, Cursor) else None,
                last=last,
                before=before if isinstance(before, Cursor) else None,
            ),
        )



from dataclasses import dataclass
from typing import Any, Dict, List, Sequence, Tuple, Type, Union, cast

import numpy as np
from hdbscan import HDBSCAN
from numpy.typing import ArrayLike
from umap import UMAP

from phoenix.datasets import Dataset

from .pointcloud import (
    Cluster,
    Coordinates,
    Coordinates2D,
    Coordinates3D,
    InferenceAttributes,
    Point,
)

MAX_UMAP_POINTS = 500
DEFAULT_MIN_CLUSTER_SIZE = 20
DEFAULT_MIN_SAMPLES = 1


@dataclass(frozen=True)
class UMAPProjector:
    hyperparameters: Dict[str, Union[int, float, str]]

    def __post__init__(self) -> None:
        if "n_neighbors" in self.hyperparameters and (
            not isinstance(self.hyperparameters["n_neighbors"], int)
            or self.hyperparameters["n_neighbors"] not in (2, 3)
        ):
            raise ValueError(
                "Projection dimensionality not supported. Must be integer value: 2 or 3 (2D/3D)."
            )

    @staticmethod
    def _move_to_center(projections: np.ndarray) -> np.ndarray:  # type: ignore
        # Calculate Center of Mass
        cm = np.sum(projections, axis=0) / projections.shape[0]
        return projections - cm  # type: ignore

    @staticmethod
    def _build_points(
        primary_projections: np.ndarray,  # type: ignore
        reference_projections: np.ndarray,  # type: ignore
        primary_dataset: Dataset,
        reference_dataset: Dataset,
        embedding_feature: str,
    ) -> Tuple[List[Point], List[Point]]:
        primary_points: List[Point] = []
        reference_points: List[Point] = []

        # Number of dimensions in projections: 2D or 3D
        N = primary_projections.shape[-1]
        c: Type[Coordinates]
        if N == 2:
            c = Coordinates2D
        elif N == 3:
            c = Coordinates3D
        else:
            raise ValueError("Projections should be done to 2D or 3D.")

        # Build primary points
        prediction_label_col = primary_dataset.get_prediction_label_column()
        actual_label_col = primary_dataset.get_actual_label_column()
        raw_text_data_col = primary_dataset.get_embedding_raw_data_column(embedding_feature)
        for i in range(len(primary_projections)):
            inference_attributes = InferenceAttributes(
                prediction_label=prediction_label_col[i],
                actual_label=actual_label_col[i],
                raw_text_data=raw_text_data_col[i],
            )
            primary_points.append(
                Point(
                    id=i,
                    coordinates=c(*[primary_projections[i][k] for k in range(N)]),
                    inference_attributes=inference_attributes,
                )
            )
        # Build reference points
        prediction_label_col = reference_dataset.get_prediction_label_column()
        actual_label_col = reference_dataset.get_actual_label_column()
        raw_text_data_col = reference_dataset.get_embedding_raw_data_column(embedding_feature)
        for i in range(len(reference_projections)):
            inference_attributes = InferenceAttributes(
                prediction_label=prediction_label_col[i],
                actual_label=actual_label_col[i],
                raw_text_data=raw_text_data_col[i],
            )
            reference_points.append(
                Point(
                    id=i + len(primary_projections),
                    coordinates=c(*[reference_projections[i][k] for k in range(N)]),
                    inference_attributes=inference_attributes,
                )
            )
        return primary_points, reference_points

    @staticmethod
    def _build_clusters(
        cluster_ids: np.ndarray,  # type: ignore
        primary_points: List[Point],
        reference_points: List[Point],
    ) -> List[Cluster]:
        unique_cluster_ids = np.unique(cluster_ids)
        # map cluster_id to point_ids inside the cluster
        map_cluster_id_point_ids: Dict[int, List[int]] = {
            id: [] for id in unique_cluster_ids if id != -1
        }
        # map cluster_id to the count of primary points in the cluster
        map_cluster_id_primary_count: Dict[int, int] = {
            id: 0 for id in unique_cluster_ids if id != -1
        }
        # map cluster_id to the count of reference points in the cluster
        map_cluster_id_reference_count: Dict[int, int] = {
            id: 0 for id in unique_cluster_ids if id != -1
        }

        primary_cluster_ids = cluster_ids[: len(primary_points)]
        reference_cluster_ids = cluster_ids[len(primary_points) :]
        # Check that there are as many coordinates as cluster IDs
        # This is a defensive test, since this should be guaranteed by UMAP & HDBSCAN libraries
        if len(reference_cluster_ids) != len(reference_points):
            raise ValueError(
                f"There should be equal number of point coordinates as cluster IDs. "
                f"len(reference_cluster_ids) = {len(reference_cluster_ids)}. "
                f"len(reference_points) = {len(reference_points)}."
            )

        for i, cluster_id in enumerate(primary_cluster_ids):
            if cluster_id == -1:  # Exclude "unknown" cluster
                continue
            map_cluster_id_point_ids[cluster_id].append(primary_points[i].id)
            map_cluster_id_primary_count[cluster_id] += 1
        for i, cluster_id in enumerate(reference_cluster_ids):

            if cluster_id == -1:  # Exclude "unknown" cluster
                continue
            map_cluster_id_point_ids[cluster_id].append(reference_points[i].id)
            map_cluster_id_reference_count[cluster_id] += 1

        clusters: List[Cluster] = []
        for cluster_id, point_ids in map_cluster_id_point_ids.items():
            primary_count = map_cluster_id_primary_count[cluster_id]
            reference_count = map_cluster_id_reference_count[cluster_id]
            purity_score = (reference_count - primary_count) / (reference_count + primary_count)
            clusters.append(Cluster(id=cluster_id, point_ids=point_ids, purity_score=purity_score))
        return clusters

    def project(
        self, primary_dataset: Dataset, reference_dataset: Dataset, embedding_feature: str
    ) -> Tuple[List[Point], List[Point], List[Cluster]]:
        # Sample down our datasets to max 2500 rows for UMAP performance
        points_per_dataset = MAX_UMAP_POINTS // 2
        sampled_primary_dataset = primary_dataset.sample(num=points_per_dataset)
        sampled_reference_dataset = reference_dataset.sample(num=MAX_UMAP_POINTS // 2)

        primary_embeddings = np.stack(
            cast(
                Sequence[ArrayLike],
                sampled_primary_dataset.get_embedding_vector_column(embedding_feature),
            )
        )
        reference_embeddings = np.stack(
            cast(
                Sequence[ArrayLike],
                sampled_reference_dataset.get_embedding_vector_column(embedding_feature),
            )
        )

        embeddings = np.concatenate([primary_embeddings, reference_embeddings])
        umap = UMAP(**self.hyperparameters)
        projections: np.ndarray = umap.fit_transform(embeddings)  # type: ignore
        projections = self._move_to_center(projections)
        # Find clusters
        hdbscan = HDBSCAN(
            min_cluster_size=DEFAULT_MIN_CLUSTER_SIZE, min_samples=DEFAULT_MIN_SAMPLES
        )
        cluster_ids: np.ndarray[Any, Any] = hdbscan.fit_predict(projections)

        primary_points, reference_points = self._build_points(
            projections[:points_per_dataset],
            projections[points_per_dataset:],
            sampled_primary_dataset,
            sampled_reference_dataset,
            embedding_feature,
        )

        clusters = self._build_clusters(cluster_ids, primary_points, reference_points)

        return primary_points, reference_points, clusters



from abc import ABC, abstractmethod
from typing import Iterable, List, Union


class ValidationError(ABC):
    def __repr__(self) -> str:
        return self.__class__.__name__

    def __str__(self) -> str:
        return self.error_message()

    @abstractmethod
    def error_message(self) -> str:
        pass


class MissingVectorColumn(ValidationError):
    """An error raised when the dataset is missing an embedding vector column specified in the
    schema"""

    def __init__(self, col: str) -> None:
        self.missing_col = col

    def error_message(self) -> str:
        return (
            f"The embedding vector column {self.missing_col} is declared in the Schema "
            "but is not found in the data."
        )


class MissingColumns(ValidationError):
    """An error raised when the dataset is missing columns specified in the schema"""

    def __init__(self, cols: Iterable[str]) -> None:
        self.missing_cols = cols

    def error_message(self) -> str:
        return (
            "The following columns are declared in the Schema "
            "but are not found in the dataframe: "
            f"{', '.join(map(str, self.missing_cols))}."
        )


class DatasetError(Exception):
    """An error raised when the dataset is invalid or incomplete"""

    def __init__(self, errors: Union[ValidationError, List[ValidationError]]):
        self.errors = errors


class MissingField(ValidationError):
    """An error raised when trying to access a field that is absent from the Schema"""

    def __init__(self, field: str) -> None:
        self.missing_field = field

    def error_message(self) -> str:
        return f"Schema is missing {self.missing_field}"


class MissingEmbeddingFeatureColumnNames(ValidationError):
    """An error raised when trying to access an EmbeddingColumnNames config that is absent
    from the schema"""

    def __init__(self, embedding_feature_name: str) -> None:
        self.embedding_feature_name = embedding_feature_name

    def error_message(self) -> str:
        return f"Schema is missing embedding_feature_column_names[{self.embedding_feature_name}]"


class MissingEmbeddingFeatureVectorColumnName(ValidationError):
    """An error raised when trying to access an EmbeddingColumnNames.vector_column_name
    that is absent from the schema"""

    def __init__(self, embedding_feature_name: str) -> None:
        self.embedding_feature_name = embedding_feature_name

    def error_message(self) -> str:
        return (
            f"Schema is missing vector_column_name of embedding_feature_column_names"
            f"[{self.embedding_feature_name}]"
        )


class MissingEmbeddingFeatureRawDataColumnName(ValidationError):
    """An error raised when trying to access an EmbeddingColumnNames.raw_data_column_name
    that is absent from the schema"""

    def __init__(self, embedding_feature_name: str) -> None:
        self.embedding_feature_name = embedding_feature_name

    def error_message(self) -> str:
        return (
            f"Schema is missing raw_data_column_name of embedding_feature_column_names"
            f"[{self.embedding_feature_name}]"
        )


class MissingEmbeddingFeatureLinkToDataColumnName(ValidationError):
    """An error raised when trying to access an EmbeddingColumnNames.link_to_data_column_name
    absent from the schema"""

    def __init__(self, embedding_feature_name: str) -> None:
        self.embedding_feature_name = embedding_feature_name

    def error_message(self) -> str:
        return (
            f"Schema is missing link_to_data_column_name of embedding_feature_column_names"
            f"[{self.embedding_feature_name}]"
        )


class SchemaError(Exception):
    """An error raised when the Schema is invalid or incomplete"""

    def __init__(self, errors: Union[ValidationError, List[ValidationError]]):
        self.errors = errors



from typing import Any

import numpy as np
import pandas as pd

from phoenix.datasets import Dataset


def euclidean_distance(
    primary: Dataset,
    reference: Dataset,
    embedding_feature_name: str,
) -> float:
    primary_vec_col: pd.Series[Any] = primary.get_embedding_vector_column(embedding_feature_name)
    reference_vec_col: pd.Series[Any] = reference.get_embedding_vector_column(
        embedding_feature_name
    )
    primary_centroid = primary_vec_col.mean()
    reference_centroid = reference_vec_col.mean()
    return np.linalg.norm(primary_centroid - reference_centroid).item()



from os import path
from pathlib import Path
from typing import Union

FilePath = Union[str, Path]


def is_url(filepath: str) -> bool:
    url_prefixes = ["http://", "https://"]
    return any([filepath.startswith(prefix) for prefix in url_prefixes])


def parse_file_format(filepath: FilePath) -> str:
    return path.splitext(filepath)[-1]


def parse_filename(filepath: FilePath) -> str:
    return path.basename(filepath)



from dataclasses import dataclass

from .dimension_data_type import DimensionDataType
from .dimension_type import DimensionType


@dataclass
class Dimension:
    name: str
    data_type: DimensionDataType
    type: DimensionType



import json
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Dict, List

MAX_UMAP_POINTS = 500


class Coordinates(ABC):
    @abstractmethod
    def get_coordinates(self) -> List[float]:
        pass


@dataclass
class Coordinates2D(Coordinates):
    x: float
    y: float

    def get_coordinates(self) -> List[float]:
        return [float(self.x), float(self.y)]


@dataclass
class Coordinates3D(Coordinates):
    x: float
    y: float
    z: float

    def get_coordinates(self) -> List[float]:
        return [float(self.x), float(self.y), float(self.z)]


@dataclass(frozen=True)
class InferenceAttributes:
    prediction_label: str
    # prediction_score: float,
    actual_label: str
    # actual_score: float,
    raw_text_data: str
    # link_to_data: str,


@dataclass(frozen=True)
class Point:
    id: int
    coordinates: Coordinates
    inference_attributes: InferenceAttributes


@dataclass(frozen=True)
class Cluster:
    id: int
    point_ids: List[int]
    purity_score: float


@dataclass(frozen=True)
class DriftPointCloud:
    primary_points: List[Point]
    reference_points: List[Point]
    clusters: List[Cluster]

    # For demo - passing to umap widget
    def to_json(self) -> str:
        primary_pts_json = self._points_to_json(self.primary_points)
        reference_pts_json = self._points_to_json(self.reference_points)
        clusters_json = self._clusters_to_json(self.clusters)

        data = {
            "primaryData": primary_pts_json,
            "referenceData": reference_pts_json,
            "clusters": clusters_json,
        }
        return json.dumps(data)

    @staticmethod
    def _points_to_json(points: List[Point]) -> List[Dict[str, Any]]:
        pts_json = []
        for point in points:
            point_json_obj = {
                "position": point.coordinates.get_coordinates(),
                "metaData": {
                    "id": int(point.id),
                    "rawTextData": [point.inference_attributes.raw_text_data],
                    "predictionLabel": point.inference_attributes.prediction_label,
                    "actualLabel": point.inference_attributes.actual_label,
                },
            }
            pts_json.append(point_json_obj)
        return pts_json

    @staticmethod
    def _clusters_to_json(clusters: List[Cluster]) -> List[Dict[str, Any]]:
        clusters_json = []
        for cluster in clusters:
            cluster_json_obj = {
                "id": int(cluster.id),
                "pointIds": cluster.point_ids,
                "purityScore": cluster.purity_score,
            }
            clusters_json.append(cluster_json_obj)
        return clusters_json



"""
Loads the compiled Javascript bundle
"""
import json
import os
from random import random
from typing import Any, Dict, List

from IPython.core.display import HTML, display  # type: ignore

import phoenix

BASE_PATH = phoenix.__path__[0]
STATIC_PATH = os.path.join(BASE_PATH, "nbextension", "static")


def load_js() -> str:
    with open(os.path.join(STATIC_PATH, "index.js"), encoding="utf-8") as f:
        return f.read()


def load_style() -> str:
    return """
        body {
            font-family: 'Roboto', sans-serif;
            }
    """


def random_point(offset: int, id: int) -> Dict[str, Any]:
    return {
        "position": [
            random() + offset,
            random() + offset,
            random() + offset,
        ],
        "metaData": {"id": id},
    }


def generate_data(length: int, offset: int, id_offset: int) -> List[Dict[str, Any]]:
    data = enumerate([None] * length)
    return list(
        map(
            lambda i: random_point(offset, id=i[0] + id_offset),
            data,
        )
    )


def demo_json() -> str:
    data = {
        "primaryData": generate_data(100, 0, 0),
        "referenceData": generate_data(100, 1, 100),
    }
    return json.dumps(data)


class UMAPWidget:
    def __init__(self, json: str) -> None:
        self.json = json

    @staticmethod
    def template(json_data: str) -> str:
        return f"""
        <html>
            <script>{load_js()}</script>
            <style>{load_style()}</style>
                <body>
                    <div id='root'>
                    </div>
                </body>
            <script>window.renderUMAPWidget({json_data});</script>
        </html>
        """

    # Temporary static json representation of UMAP Drift data

    def show(self) -> None:
        display(HTML(self.template(self.json)))



import logging
import os
import sys
import uuid
from typing import Any, Literal, Optional, Union

from pandas import DataFrame, Series, read_parquet

from phoenix.config import dataset_dir
from phoenix.utils import FilePath

from . import errors as err
from .schema import EmbeddingColumnNames, Schema
from .validation import validate_dataset_inputs

SUPPORTED_URL_FORMATS = sorted(["hdf", "csv"])

logger = logging.getLogger(__name__)
if hasattr(sys, "ps1"):
    # for python interactive mode
    log_handler = logging.StreamHandler(sys.stdout)
    log_handler.setLevel(logging.INFO)
    logger.addHandler(log_handler)
    logger.setLevel(logging.INFO)

ParquetEngine = Literal["auto", "fastparquet", "pyarrow"]


class Dataset:
    """
    A dataset represents data for a set of inferences. It is represented as a dataframe + schema
    """

    _data_file_name: str = "data.parquet"
    _schema_file_name: str = "schema.json"
    _is_persisted: bool = False

    def __init__(
        self,
        dataframe: DataFrame,
        schema: Schema,
        name: Optional[str] = None,
        persist_to_disc: bool = True,
    ):
        errors = validate_dataset_inputs(
            dataframe=dataframe,
            schema=schema,
        )
        if errors:
            for e in errors:
                logger.error(e)
            raise err.DatasetError(errors)
        parsed_dataframe = self._parse_dataframe(dataframe, schema)

        self.__dataframe: DataFrame = parsed_dataframe
        self.__schema: Schema = schema
        self.__name: str = name if name is not None else f"""dataset_{str(uuid.uuid4())}"""
        self.__directory: str = os.path.join(dataset_dir, self.name)

        # Sync the dataset to disc so that the server can pick up the data
        if persist_to_disc:
            self.to_disc()
        else:
            # Assume that the dataset is already persisted to disc
            self._is_persisted: bool = True

        self.to_disc()
        logger.info(f"""Dataset: {self.__name} initialized""")

    @property
    def dataframe(self) -> DataFrame:
        return self.__dataframe

    @property
    def schema(self) -> "Schema":
        return self.__schema

    @property
    def name(self) -> str:
        return self.__name

    @property
    def is_persisted(self) -> bool:
        return self._is_persisted

    @property
    def directory(self) -> str:
        """The directory under which the dataset metadata is stored"""
        return self.__directory

    def head(self, num_rows: Optional[int] = 5) -> DataFrame:
        num_rows = 5 if num_rows is None else num_rows
        return self.dataframe.head(num_rows)

    def get_column(self, col_name: str) -> "Union[Series[int], Series[float], Series[str]]":
        return self.dataframe[col_name]

    def sample(self, num: int) -> "Dataset":
        sampled_dataframe = self.dataframe.sample(n=num, ignore_index=True)
        return Dataset(sampled_dataframe, self.schema, f"""{self.name}_sample_{num}""")

    def get_prediction_label_column(
        self,
    ) -> "Series[str]":
        if self.schema.prediction_label_column_name is None:
            raise err.SchemaError(err.MissingField("prediction_label_column_name"))
        return self.dataframe[self.schema.prediction_label_column_name]

    def get_prediction_score_column(
        self,
    ) -> "Series[float]":
        if self.schema.prediction_score_column_name is None:
            raise err.SchemaError(err.MissingField("prediction_score_column_name"))
        return self.dataframe[self.schema.prediction_score_column_name]

    def get_actual_label_column(self) -> "Series[str]":
        if self.schema.actual_label_column_name is None:
            raise err.SchemaError(err.MissingField("actual_label_column_name"))
        return self.dataframe[self.schema.actual_label_column_name]

    def get_actual_score_column(self) -> "Union[Series[int], Series[float], Series[str]]":
        if self.schema.actual_score_column_name is None:
            raise err.SchemaError(err.MissingField("actual_score_column_name"))
        return self.dataframe[self.schema.actual_score_column_name]

    def _get_embedding_feature_column_names(
        self, embedding_feature_name: str
    ) -> EmbeddingColumnNames:
        if self.schema.embedding_feature_column_names is None:
            raise err.SchemaError(err.MissingField("embedding_feature_column_names"))
        embedding_feature_column_names = self.schema.embedding_feature_column_names
        if (
            embedding_feature_name not in embedding_feature_column_names
            or embedding_feature_column_names[embedding_feature_name] is None
        ):
            raise err.SchemaError(err.MissingEmbeddingFeatureColumnNames(embedding_feature_name))
        return embedding_feature_column_names[embedding_feature_name]

    # TODO(mikeldking): add strong vector type
    def get_embedding_vector_column(self, embedding_feature_name: str) -> "Series[Any]":
        column_names = self._get_embedding_feature_column_names(embedding_feature_name)
        if column_names.vector_column_name is None:
            raise err.SchemaError(
                err.MissingEmbeddingFeatureVectorColumnName(embedding_feature_name)
            )
        vector_column = self.dataframe[column_names.vector_column_name]
        return vector_column

    def get_embedding_raw_data_column(self, embedding_feature_name: str) -> "Series[str]":
        column_names = self._get_embedding_feature_column_names(embedding_feature_name)
        if column_names.raw_data_column_name is None:
            raise err.SchemaError(
                err.MissingEmbeddingFeatureRawDataColumnName(embedding_feature_name)
            )
        return self.dataframe[column_names.raw_data_column_name]

    def get_embedding_link_to_data_column(self, embedding_feature_name: str) -> "Series[str]":
        column_names = self._get_embedding_feature_column_names(embedding_feature_name)
        if column_names.link_to_data_column_name is None:
            raise err.SchemaError(
                err.MissingEmbeddingFeatureLinkToDataColumnName(embedding_feature_name)
            )
        return self.dataframe[column_names.link_to_data_column_name]

    @classmethod
    def from_dataframe(
        cls, dataframe: DataFrame, schema: Schema, name: Optional[str] = None
    ) -> "Dataset":
        return cls(dataframe, schema, name)

    @classmethod
    def from_parquet(
        cls,
        filepath: FilePath,
        schema: Schema,
        name: Optional[str] = None,
        engine: ParquetEngine = "pyarrow",
    ) -> "Dataset":
        return cls(read_parquet(filepath, engine=engine), schema, name)

    @classmethod
    def from_name(cls, name: str) -> "Dataset":
        """Retrieves a dataset by name from the file system"""
        directory = os.path.join(dataset_dir, name)
        df = read_parquet(os.path.join(directory, cls._data_file_name))
        with open(os.path.join(directory, cls._schema_file_name)) as schema_file:
            schema_json = schema_file.read()
        schema = Schema.from_json(schema_json)
        return cls(df, schema, name, persist_to_disc=False)

    @staticmethod
    def _parse_dataframe(dataframe: DataFrame, schema: Schema) -> DataFrame:
        schema_cols = [
            schema.timestamp_column_name,
            schema.prediction_label_column_name,
            schema.prediction_score_column_name,
            schema.actual_label_column_name,
            schema.actual_score_column_name,
        ]
        # Append the feature column names to the columns if present
        if schema.feature_column_names is not None:
            schema_cols += schema.feature_column_names

        if schema.embedding_feature_column_names is not None:
            for emb_feat_cols in schema.embedding_feature_column_names.values():
                schema_cols.append(emb_feat_cols.vector_column_name)
                if emb_feat_cols.raw_data_column_name:
                    schema_cols.append(emb_feat_cols.raw_data_column_name)
                if emb_feat_cols.link_to_data_column_name:
                    schema_cols.append(emb_feat_cols.link_to_data_column_name)

        drop_cols = [col for col in dataframe.columns if col not in schema_cols]
        return dataframe.drop(columns=drop_cols)

    def to_disc(self) -> None:
        """writes the data and schema to disc"""

        if self._is_persisted:
            logger.info("Dataset already persisted")
            return

        directory = self.directory
        if not os.path.isdir(directory):
            os.makedirs(directory)

        self.dataframe.to_parquet(os.path.join(directory, self._data_file_name))
        schema_json_data = self.schema.to_json()
        with open(os.path.join(directory, self._schema_file_name), "w+") as schema_file:
            schema_file.write(schema_json_data)

        # set the persisted flag so that we don't have to perform this operation again
        self._is_persisted = True
        logger.info(f"Dataset info written to '{directory}'")


def show_progress(block_num: int, block_size: int, total_size: int) -> None:
    progress = round(block_num * block_size / total_size * 100, 2)
    print("[" + int(progress) * "=" + (100 - int(progress)) * " " + f"] {progress}%", end="\r")
